MODULE ExMathLib0;

 (*
    ExMathLib0 - Extended precision math library.       
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT LR := LRealStr, LM := LRealMath, X := ExNumbers;

CONST
  (* DegRadType definitions *)
  Degrees  * = 0;
  Radians  * = 1;
  Gradians * = 2;

TYPE
  DegRadType * = SHORTINT;

VAR
  FromDegrees : X.ExNumType;
  ToDegrees   : X.ExNumType;
  ToGradians  : X.ExNumType;
  FromGradians: X.ExNumType;
  Fact500     : X.ExNumType;
  Fact1000    : X.ExNumType;
  Fact2000    : X.ExNumType;
  Fact3000    : X.ExNumType;
  AngleUnit*  : DegRadType;
  Seed        : X.ExNumType;


PROCEDURE ExNumToLongReal*(x : X.ExNumType) : LONGREAL;
VAR
  Num : LONGREAL;
  res : SHORTINT;
  Str : ARRAY 81 OF CHAR;
BEGIN
  (* Convert ExNum into LONGREAL via a string *)
  X.ExNumToStr(x, 0, 0, Str, FALSE);
  LR.StrToReal(Str, Num, res);
  IF res=LR.strAllRight THEN RETURN Num
  ELSE RETURN 0.0
  END
END ExNumToLongReal;


PROCEDURE LongRealToExNum*(x : LONGREAL; VAR Result : X.ExNumType);
VAR
  Str : ARRAY 128 OF CHAR;
BEGIN
  (* Convert LONGREAL into an ExNum via a string *)
  LR.RealToFloat(x, 52, Str);
  X.StrToExNum(Str, Result)
END LongRealToExNum;


PROCEDURE xtoi*(VAR Result : X.ExNumType; x : X.ExNumType; i : LONGINT);
(* From Knuth, slightly altered : p442, The Art Of Computer Programming, Vol 2 *)
VAR
  Y : X.ExNumType;
  negative : BOOLEAN;
BEGIN
  Y := X.Ex1;
  negative := i < 0;
  i := ABS(i);
  LOOP
    IF ODD(i) THEN X.ExMult(Y, Y, x) END;
    i := i DIV 2;
    IF i = 0 THEN EXIT END;
    X.ExMult(x, x, x);
  END;
  IF negative THEN
    X.ExDiv(Result, X.Ex1, Y);
  ELSE
    Result := Y;
  END;
END xtoi;


PROCEDURE Root *(VAR Result : X.ExNumType;
                    x      : X.ExNumType;
                    i      : LONGINT);
(* Use iterative solution of a general root equation *)
VAR
  y, yp, f, g, t : X.ExNumType;
  iteration : INTEGER;
  root : LONGREAL;
  negate : BOOLEAN;
BEGIN
  IF ((x.Sign = X.negative) & ~ODD(i)) OR (i < 2) THEN
    X.ExStatus := X.IllegalNumber;
    Result := X.Ex0;
  ELSIF X.IsZero(x) THEN
    Result := x;
  ELSE
    (* handle negative roots *)
    IF x.Sign = X.negative THEN X.ExAbs(x); negate := TRUE
    ELSE negate := FALSE
    END;

    (* estimate of the ith root *)
    root := 1.0 / i;
    LongRealToExNum(LM.power(ExNumToLongReal(x), root), yp);
    X.ExNumb(i, 0, 0, f);    (* i *)
    X.ExNumb(i-1, 0, 0, g);  (* i - 1 *)

    (* calculate the root *)
    iteration := 4;
    LOOP
      (* y := 1/i * (yp * (i-1) + x / yp^(i-1)) *)
      xtoi(t, yp, i-1);       (* yp**(i-1) *)
      X.ExMult(y, t, yp);     (* yp**i *)
      X.ExMult(y, y, g);      (* yp**i * (i-1) *)
      X.ExAdd(y, y, x);       (* yp**i * (i-1) + x *)
      X.ExMult(t, t, f);      (* yp**(i-1) * i *)
      X.ExDiv(y, y, t);
      IF (X.ExCompare(y, yp) = X.ExEqual) OR (iteration = 0) THEN EXIT END;
      DEC(iteration);
      yp := y;
    END;

    (* adjust the number's sign *)
    Result := y;
    IF negate THEN X.ExChgSign(Result) END;
  END;
END Root;


PROCEDURE FromRadians*(VAR radianAngle : X.ExNumType);
(* Convert a radian measure into degrees *)
BEGIN
  IF AngleUnit=Degrees THEN
    X.ExMult(radianAngle, ToDegrees, radianAngle)
  ELSIF AngleUnit=Gradians THEN
    X.ExMult(radianAngle, ToGradians, radianAngle)
  END
END FromRadians;


PROCEDURE ToRadians*(VAR radianAngle : X.ExNumType);
(* Convert an angular measure into radians *)
BEGIN
  IF AngleUnit=Degrees THEN
    X.ExMult(radianAngle, FromDegrees, radianAngle)
  ELSIF AngleUnit=Gradians THEN
    X.ExMult(radianAngle, FromGradians, radianAngle)
  END
END ToRadians;


PROCEDURE sqrtX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN
  Root(Result, x, 2);
END sqrtX;


PROCEDURE lnX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN
  LongRealToExNum(LM.ln(ExNumToLongReal(x)), Result);
END lnX;


PROCEDURE logX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN
  LongRealToExNum(LM.log(ExNumToLongReal(x), 10), Result);
END logX;


PROCEDURE factorial(VAR prevn, currentn : LONGINT;
                    VAR PrevFact, Result : X.ExNumType);
(* Implements an incremental factorial using a previously calculated value. *)
VAR
  i : LONGINT;
BEGIN
  FOR i := prevn+1 TO currentn DO
    (* PrevFact := PrevFact * i; *)
    X.ExNumb(i, 0, 0, Result);
    X.ExMult(PrevFact, PrevFact, Result);
  END;
  prevn := currentn;
  Result := PrevFact;
END factorial;


PROCEDURE factorialX*(VAR Result : X.ExNumType; n : LONGINT);
CONST
  MaxFactorial = 3249;
VAR
  fact : LONGINT;
  prev : X.ExNumType;
BEGIN
  IF (n < 0) OR (n > MaxFactorial) THEN
    X.ExStatus := X.IllegalNumber;
    Result := X.Ex0;
    RETURN;
  END;
  IF    n < 500  THEN prev := X.Ex1;      fact := 0
  ELSIF n < 1000 THEN prev := Fact500;  fact := 500
  ELSIF n < 2000 THEN prev := Fact1000; fact := 1000
  ELSIF n < 3000 THEN prev := Fact2000; fact := 2000
  ELSE                prev := Fact3000; fact := 3000
  END;
  factorial(fact, n, prev, Result);
END factorialX;


PROCEDURE permutations * (VAR Result : X.ExNumType; n, r: X.ExNumType);
(* Return the number of permutations of n different objects
   taken r at a time.
 *)
VAR ni, ri: LONGINT;
BEGIN
  ni:=X.ExToLongInt(n); ri:=X.ExToLongInt(r); n:=X.Ex1;
  factorial(ri, ni, n, Result)
END permutations;


PROCEDURE combinations * (VAR Result : X.ExNumType; n, r: X.ExNumType);
VAR ni, ri: LONGINT;
BEGIN
  ni:=X.ExToLongInt(n); ri:=X.ExToLongInt(r); n:=X.Ex1;
  factorial(ri, ni, n, r);         (* r=n!/(n-r)! *)
  factorialX(n, X.ExToLongInt(r)); (* n=r! *)
  X.ExDiv(Result, r, n)            (* Result=n!/(r!(n-r)!) *)
END combinations;


PROCEDURE expX*(VAR Result : X.ExNumType; x : X.ExNumType);
VAR
  xPower : LONGREAL;
BEGIN
  xPower := ExNumToLongReal(x);
  X.ExFrac(x);
  IF (ABS(xPower) < MAX(LONGINT)) & X.IsZero(x) THEN
    xtoi(Result, X.e, ENTIER(xPower));
  ELSE
    LongRealToExNum(LM.exp(xPower), Result);
  END;
END expX;


PROCEDURE powerX*(VAR Result : X.ExNumType; x, y : X.ExNumType);
VAR
  yPower : LONGREAL;
BEGIN
  yPower := ExNumToLongReal(y);
  X.ExFrac(y);
  IF (ABS(yPower) < MAX(LONGINT)) & X.IsZero(y) THEN
    xtoi(Result, x, ENTIER(yPower));
  ELSE
    LongRealToExNum(LM.power(ExNumToLongReal(x), yPower),Result);
  END;
END powerX;


PROCEDURE rootX*(VAR Result : X.ExNumType; x, y : X.ExNumType);
VAR
  yRoot : LONGREAL;
BEGIN
  yRoot := ExNumToLongReal(y);
  X.ExFrac(y);
  IF (ABS(yRoot) < MAX(LONGINT)) & X.IsZero(y) THEN
    Root(Result, x, ENTIER(yRoot));
  ELSE
    yRoot := 1.0 / yRoot;
    LongRealToExNum(LM.power(ExNumToLongReal(x), yRoot),Result);
  END;
END rootX;


PROCEDURE random*(VAR Result : X.ExNumType);
(* Return a random number between 0 and 1 *)
VAR t, five: X.ExNumType;
BEGIN
  X.ExNumb(5, 0, 0, five);
  X.ExAdd(Result, Seed, X.pi);
  lnX(t, Result); X.ExMult(t, t, five);
  expX(t, t); Result:=t; X.ExTrunc(t);
  X.ExSub(Seed, Result, t);
  Result:=Seed
END random;


PROCEDURE sinX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN ToRadians(x);
  LongRealToExNum(LM.sin(ExNumToLongReal(x)), Result);
END sinX;


PROCEDURE cosX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN ToRadians(x);
  LongRealToExNum(LM.cos(ExNumToLongReal(x)), Result);
END cosX;


PROCEDURE tanX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN ToRadians(x);
  LongRealToExNum(LM.tan(ExNumToLongReal(x)), Result);
END tanX;


PROCEDURE arctanX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN
  LongRealToExNum(LM.arctan(ExNumToLongReal(x)), Result);
  FromRadians(Result)
END arctanX;


PROCEDURE arctan2X*(VAR Result : X.ExNumType; x, y : X.ExNumType);
VAR half, t : X.ExNumType;
BEGIN
  IF X.IsZero(x) THEN X.ExNumb(0, 5, 0, half);  (* 0.5 *)
    IF X.IsZero(y) THEN Result := X.Ex0 ELSE X.ExMult(Result, X.pi, half) END
  ELSIF X.IsZero(y) THEN
    IF x.Sign = X.positive THEN Result := X.Ex0 ELSE Result := X.pi END
  ELSE X.ExDiv(t, y, x); X.ExAbs(t); arctanX(t, t);
    IF x.Sign = X.positive THEN Result := t;
      IF y.Sign = X.negative THEN X.ExChgSign(Result) (* x > 0, y < 0 *) END
    ELSE
      IF y.Sign = X.positive THEN X.ExSub(Result, X.pi, t) (* x < 0, y > 0 *)
      ELSE X.ExAdd(Result, X.pi, t) (* x,y < 0 *)
      END
    END
  END;
  FromRadians(Result)
END arctan2X;


PROCEDURE coshX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN
  LongRealToExNum(LM.cosh(ExNumToLongReal(x)), Result);
END coshX;


PROCEDURE sinhX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN
  LongRealToExNum(LM.sinh(ExNumToLongReal(x)), Result);
END sinhX;


PROCEDURE tanhX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN
  LongRealToExNum(LM.tanh(ExNumToLongReal(x)), Result);
END tanhX;


PROCEDURE arccoshX*(VAR Result : X.ExNumType; x : X.ExNumType);
VAR
  Temp : X.ExNumType;
BEGIN
  (* Result = ln(x + sqrt(x*x - 1)) *)
  X.ExMult(Temp, x, x);
  X.ExSub(Temp, Temp, X.Ex1);
  sqrtX(Temp, Temp);
  X.ExAdd(Temp, x, Temp);
  lnX(Result, Temp);
END arccoshX;


PROCEDURE arcsinhX*(VAR Result : X.ExNumType; x : X.ExNumType);
VAR
  Temp : X.ExNumType;
BEGIN
  (* Result = ln(x + sqrt(x*x + 1)) *)
  X.ExMult(Temp, x, x);
  X.ExAdd(Temp, Temp, X.Ex1);
  sqrtX(Temp, Temp);
  X.ExAdd(Temp, x, Temp);
  lnX(Result, Temp);
END arcsinhX;


PROCEDURE arctanhX*(VAR Result : X.ExNumType; x : X.ExNumType);
VAR
  Temp, Temp2 : X.ExNumType;
BEGIN
  (* Result = ln((1 + x) / (1 - x)) / 2 *)
  X.ExAdd(Temp, X.Ex1, x);
  X.ExSub(Temp2, X.Ex1, x);
  X.ExDiv(Temp, Temp, Temp2);
  lnX(Result, Temp);
  X.ExNumb(0, 5, 0, Temp);
  X.ExMult(Result, Result, Temp);
END arctanhX;


PROCEDURE arcsinX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN
  LongRealToExNum(LM.arcsin(ExNumToLongReal(x)), Result);
  FromRadians(Result)
END arcsinX;


PROCEDURE arccosX*(VAR Result : X.ExNumType; x : X.ExNumType);
BEGIN
  (* Replacement algorithm *)
  LongRealToExNum(LM.arccos(ExNumToLongReal(x)), Result);
  FromRadians(Result)
END arccosX;


BEGIN
  (* Initialize a few internal conversion constants *)
  AngleUnit:=Degrees;   (* default angular units *)
  X.StrToExNum("5.729577951308232087679815481410517033240547246656420E+1",ToDegrees);
  X.StrToExNum("1.745329251994329576923690768488612713442871888541727E-2",FromDegrees);
  X.StrToExNum("6.366197723675813430755350534900574481378385829618240E+1",ToGradians);
  X.StrToExNum("1.570796326794896619231321691639751442098584699687555E-2",FromGradians);

  (* Speed up very large factorials *)
  X.StrToExNum("1.220136825991110068701238785423046926253574342803193E+1134",Fact500);
  X.StrToExNum("4.023872600770937735437024339230039857193748642107146E+2567",Fact1000);
  X.StrToExNum("3.316275092450633241175393380576324038281117208105780E+5735",Fact2000);
  X.StrToExNum("4.149359603437854085556867093086612170951119194931810E+9130",Fact3000);

  (* Random number generator *)
  X.ExNumb(4, 0, 0, Seed)
END ExMathLib0.
